/*
 * Copyright (C) 2018 Purism SPC
 *
 * SPDX-License-Identifier: LGPL-2.1+
 */

#include "config.h"
#include "hdy-fold.h"

/**
 * SECTION:hdy-fold
 * @short_description: Element folding states.
 * @title: HdyFold
 */

/**
 * HdyFold:
 * @HDY_FOLD_UNFOLDED: The element isn't folded
 * @HDY_FOLD_FOLDED: The element is folded
 *
 * Represents the fold of widgets and other objects which can be switched
 * between folded and unfolded state on the fly, like HdyLeaflet.
 */
