.seahorse-sidebar-item-header {
  margin-top: 6px;
  font-weight: bold;
}

.seahorse-sidebar-item {
  margin: 0 3px 0 12px;
}

.seahorse-sidebar-item > label {
  margin: 3px 0 3px 0;
}

.seahorse-sidebar-item > button {
  padding: 0;
  margin: 0;
}

.seahorse-item-listbox {
  border: 1px solid @borders;
  min-width: 400px;
}

.seahorse-item-listbox > row {
  border-bottom: 1px groove @borders;
}

.seahorse-item-listbox > row:last-child {
  border-bottom: 0;
}

.seahorse-item-listbox-row {
  margin: 12px;
}

.seahorse-item-listbox-row > .seahorse-item-listbox-row-description {
  font-size: small;
}

.new-item-list {
    background-color: transparent;
}

.pane-separator {
    border-width: 1px 1px 0 0;
    border-style: solid;
    border-color: shade (@theme_bg_color, 0.66);
}

levelbar .strength-weak {
  background-color: #cc0000;
  border-color: #cc0000;
}

levelbar .strength-low {
  background-color: #f5ce00;
  border-color: #f5ce00;
}

levelbar .strength-medium,
levelbar .strength-good,
levelbar .strength-high {
  background-color: #73d216;
  border-color: #73d216;
}
