hdyleaflet > dimming {
  background: rgba(0, 0, 0, 0.12);
}

hdyleaflet > shadow {
  min-width: 56px;
  min-height: 56px;
}

hdyleaflet > border {
  min-width: 1px;
  min-height: 1px;
  background: rgba(0, 0, 0, 0.05);
}

hdyleaflet > shadow.left {
  background-image:
    linear-gradient(to right,
                    alpha(#000, 0.05),
                    alpha(#000, 0.01) 40px,
                    alpha(#000, 0) 56px),
    linear-gradient(to right,
                    alpha(#000, 0.03),
                    alpha(#000, 0.01) 7px,
                    alpha(#000, 0) 24px);
}

hdyleaflet > shadow.right {
  background-image:
    linear-gradient(to left,
                    alpha(#000, 0.05),
                    alpha(#000, 0.01) 40px,
                    alpha(#000, 0) 56px),
    linear-gradient(to left,
                    alpha(#000, 0.03),
                    alpha(#000, 0.01) 7px,
                    alpha(#000, 0) 24px);

}

hdyleaflet > shadow.up {
  background-image:
    linear-gradient(to bottom,
                    alpha(#000, 0.05),
                    alpha(#000, 0.01) 40px,
                    alpha(#000, 0) 56px),
    linear-gradient(to bottom,
                    alpha(#000, 0.03),
                    alpha(#000, 0.01) 7px,
                    alpha(#000, 0) 24px);
}

hdyleaflet > shadow.down {
  background-image:
    linear-gradient(to top,
                    alpha(#000, 0.05),
                    alpha(#000, 0.01) 40px,
                    alpha(#000, 0) 56px),
    linear-gradient(to top,
                    alpha(#000, 0.03),
                    alpha(#000, 0.01) 7px,
                    alpha(#000, 0) 24px);
}
